int biggest2thPotency(int number) {
  int res = 1;

  while (res <= number) {
    res <<= 1;  // multiply res with 2 (res << 1) as long as res isn't bigger than the number,
  }
  // so that we have the nearest (bigger) 2th potency.
  res >>= 1;  // now divide the result once so that we have the nearest (but smaller or equal) 2th potency
  return res;
}
bool checkCommitID(int commit_id, int searched_bit) {
  // a function that checks if the searched_bit was ever set
  // Technically: the commit_id is an Integer which is the sum of other numbers (related to the ext4 permissions). List of those numbers:
  // *** 1: WiFi             ***
  // *** 2: Alternative WiFi ***
  // *** 4: Code             ***
  // *** 8: Friendly name    ***

  int b2thP;
  while (commit_id > 0) {
    b2thP = biggest2thPotency(commit_id);
    if (b2thP == searched_bit) {
      return true;
    }
    commit_id -= b2thP;
  }
  return false;
}
bool hasWiFi(int commit_id)            {
  return checkCommitID(commit_id, 1);
}
bool hasAlternativeWiFi(int commit_id) {
  return checkCommitID(commit_id, 2);
}
bool hasCode(int commit_id)            {
  return checkCommitID(commit_id, 4);
}
bool hasFriendlyName(int commit_id)    {
  return checkCommitID(commit_id, 8);
}

void setFriendlyName(String name) {
  // get the configuration
  Config config; EEPROM.get(configAddress, config);

  // convert the String to a char array
  int nameLength = name.length();
  char nameArray[21];
  for (int i = 0; i <= 20; i++) {
    if (nameLength >= i) {
      nameArray[i] = name[i];
    } else {
      nameArray[i] = (char) 0;
    }
  }

  // get the commit id
  int commit_id; EEPROM.get(0, commit_id);

  if (hasFriendlyName(commit_id)) {
    strncpy(config.friendly_name, nameArray, 21);
  }
  else                           {
    strncpy(config.friendly_name, nameArray, 21);
    EEPROM.put(0, commit_id + 8);
  }

  // save the config
  EEPROM.put(configAddress, config);
  EEPROM.commit();
}
void setCode(String code) {
  // get the configuration
  Config config; EEPROM.get(configAddress, config);

  // convert the String to a char array
  int codeLength = code.length();
  char codeArray[9];
  for (int i = 0; i <= 8; i++) {
    if (codeLength >= i) {
      codeArray[i] = code[i];
    } else {
      codeArray[i] = (char) 0;
    }
  }

  // get the commit id
  int commit_id; EEPROM.get(0, commit_id);

  if (hasCode(commit_id)) {
    strncpy(config.code, codeArray, 9);
  }
  else                   {
    strncpy(config.code, codeArray, 9);
    EEPROM.put(0, commit_id + 4);
  }

  // save the config
  EEPROM.put(configAddress, config);
  EEPROM.commit();
}
void setWiFi(String ssid, String password) {
  // get the configuration
  WiFiConfig config; EEPROM.get(wiFiConfigAddress, config);

  // convert the Strings to char arrays
  int ssidLength = ssid.length(); int pwdLength = password.length();
  char ssidArray[33];
  char pwdArray[64];
  for (int i = 0; i <= 31; i++) {
    if (ssidLength >= i) {
      ssidArray[i] = ssid[i];
    } else {
      ssidArray[i] = (char) 0;
    }
  }
  for (int i = 0; i <= 62; i++) {
    if (pwdLength >= i) {
      pwdArray[i] = password[i];
    } else {
      pwdArray[i] = (char) 0;
    }
  }

  // get the commit id
  int commit_id; EEPROM.get(0, commit_id);

  if (hasWiFi(commit_id)) {
    strncpy(config.ssid, ssidArray, 33);
    strncpy(config.pwd, pwdArray, 64);
  }
  else                   {
    strncpy(config.ssid, ssidArray, 33);
    strncpy(config.pwd, pwdArray, 64);
    EEPROM.put(0, commit_id + 1);
  }

  // save the config
  EEPROM.put(wiFiConfigAddress, config);
  EEPROM.commit();

  // reboot to go into the wifi
  ESP.restart();
}
void setAlternativeWiFi(String ssid, String password) {
  // get the configuration
  WiFiConfig config; EEPROM.get(wiFiConfigAddress, config);

  // convert the Strings to char arrays
  int ssidLength = ssid.length(); int pwdLength = password.length();
  char ssidArray[33];
  char pwdArray[64];
  for (int i = 0; i <= 31; i++) {
    if (ssidLength >= i) {
      ssidArray[i] = ssid[i];
    } else {
      ssidArray[i] = (char) 0;
    }
  }
  for (int i = 0; i <= 62; i++) {
    if (pwdLength >= i) {
      pwdArray[i] = password[i];
    } else {
      pwdArray[i] = (char) 0;
    }
  }

  // get the commit id
  int commit_id; EEPROM.get(0, commit_id);

  if (hasAlternativeWiFi(commit_id)) {
    strncpy(config.alternative_ssid, ssidArray, 33);
    strncpy(config.alternative_pwd, pwdArray, 64);
  }
  else                   {
    strncpy(config.alternative_ssid, ssidArray, 33);
    strncpy(config.alternative_pwd, pwdArray, 64);
    EEPROM.put(0, commit_id + 2);
  }

  // save the config
  EEPROM.put(wiFiConfigAddress, config);
  EEPROM.commit();

  // reboot to go into the wifi
  ESP.restart();
}
String getFriendlyName() {
  Config config; EEPROM.get(configAddress, config);
  int commitID; EEPROM.get(0, commitID);
  if (hasFriendlyName(commitID)) {
    return (String) config.friendly_name;
  }
  else {
    setFriendlyName("ESafeP #1");  // set a default
    return getFriendlyName();
  }
}
String getCode() {
  Config config; EEPROM.get(configAddress, config);
  int commitID; EEPROM.get(0, commitID);
  if (hasCode(commitID)) {
    return (String) config.code;
  }
  else {
    setCode("12345678");  // set a default
    return getCode();
  }
}
String getWiFiSSID() {
  WiFiConfig config; EEPROM.get(wiFiConfigAddress, config);
  int commitID; EEPROM.get(0, commitID);
  if (hasWiFi(commitID)) {
    return (String) config.ssid;
  }
  else {
    return "";
  }
}
String getWiFiPassword() {
  WiFiConfig config; EEPROM.get(wiFiConfigAddress, config);
  int commitID; EEPROM.get(0, commitID);
  if (hasWiFi(commitID)) {
    return (String) config.pwd;
  }
  else {
    return "";
  }
}
String getAlternativeWiFiSSID() {
  WiFiConfig config; EEPROM.get(wiFiConfigAddress, config);
  int commitID; EEPROM.get(0, commitID);
  if (hasAlternativeWiFi(commitID)) {
    return (String) config.alternative_ssid;
  }
  else {
    return "";
  }
}
String getAlternativeWiFiPassword() {
  WiFiConfig config; EEPROM.get(wiFiConfigAddress, config);
  int commitID; EEPROM.get(0, commitID);
  if (hasAlternativeWiFi(commitID)) {
    return (String) config.alternative_pwd;
  }
  else {
    return "";
  }
}
void reset() {
  EEPROM.begin(4096);

  for (int i = 0 ; i <= 4096 ; i++) {
    EEPROM.write(i, 0);
  }

  EEPROM.commit();
}
void open() {
  doorServo.write(servo_open_angle);
  safeOpened = true;

  // if this is not done, the ESP32Servo library would try to set the servo angle to 1 degree all the time. So this resets the connection to the servo motor
  delay(400);
  doorServo.detach();
  doorServo.attach(doorServoPin);
}
void close() {
  doorServo.write(servo_closed_angle);
  safeOpened = false;

  // if this is not done, the ESP32Servo library would try to set the servo angle to 1 degree all the time. So this resets the connection to the servo motor
  delay(400);
  doorServo.detach();
  doorServo.attach(doorServoPin);
}
bool isIn(String target, String list[], int listlength) {
  for (int i = 0; i < listlength; i++) {
    //list[i].toCharArray(listElmnt, list[i].length());
    if (list[i] == target) {
      return true;
    }
  }
  return false;
}
void blinkLED(int pin, int length) {
  digitalWrite(pin, HIGH);
  delay(length);
  digitalWrite(pin, LOW);
}

void pinSetup() {
  pinMode(keyboard3x4_column1, OUTPUT);  // The I/O Pins of the 3x4 Keyboard
  pinMode(keyboard3x4_column2, OUTPUT);  // The I/O Pins of the 3x4 Keyboard
  pinMode(keyboard3x4_column3, OUTPUT);  // The I/O Pins of the 3x4 Keyboard
  pinMode(keyboard3x4_row1, INPUT);      // The I/O Pins of the 3x4 Keyboard
  pinMode(keyboard3x4_row2, INPUT);      // The I/O Pins of the 3x4 Keyboard
  pinMode(keyboard3x4_row3, INPUT);      // The I/O Pins of the 3x4 Keyboard
  pinMode(keyboard3x4_row4, INPUT);      // The I/O Pins of the 3x4 Keyboard

  pinMode(readyForInputLED, OUTPUT);
  pinMode(wrongCodeLED, OUTPUT);
  pinMode(codeChangeLED, OUTPUT);
  pinMode(newCodeLED, OUTPUT);
  pinMode(newCodeRepeatLED, OUTPUT);
  pinMode(invalidCodeLED, OUTPUT);
  pinMode(typeOldCodeLED, OUTPUT);
}
